import unittest
import jax
import jax.numpy as jnp
import rodeo.kalmantv.standard as ktv
import utils


# --- kalmantv.predict ---------------------------------------------------------


class TestKalmanTVGM(unittest.TestCase):
    """
    Test if KalmanTV gives the same results as Gaussian Markov process.

    """
    setUp = utils.kalman_setup

    def test_predict(self):
        mean_state_past, var_state_past, mean_state_pred1, var_state_pred1, _, _ = utils.test_filter(self)
        mean_state_pred2, var_state_pred2 = jax.jit(ktv.predict)(
            mean_state_past=mean_state_past,
            var_state_past=var_state_past,
            mean_state=self.mean_state[1],
            wgt_state=self.wgt_state[0],
            var_state=self.var_state[1]
        )

        self.assertAlmostEqual(utils.rel_err(
            mean_state_pred1, mean_state_pred2), 0.0)
        self.assertAlmostEqual(utils.rel_err(
            var_state_pred1, var_state_pred2), 0.0)
        
        
    def test_update(self):
        _, _, mean_state_pred, var_state_pred,  mean_state_filt1, var_state_filt1 = utils.test_filter(self)
        mean_state_filt2, var_state_filt2 = jax.jit(ktv.update)(
            mean_state_pred=mean_state_pred,
            var_state_pred=var_state_pred,
            x_meas=self.x_meas[1],
            mean_meas=self.mean_meas[1],
            wgt_meas=self.wgt_meas[1],
            var_meas=self.var_meas[1]
        )

        self.assertAlmostEqual(utils.rel_err(
            mean_state_filt1, mean_state_filt2), 0.0)
        self.assertAlmostEqual(utils.rel_err(
            var_state_filt1, var_state_filt2), 0.0)
        

    def test_filter(self):
        mean_state_past, var_state_past, mean_state_pred1, \
            var_state_pred1, mean_state_filt1, var_state_filt1 = utils.test_filter(self)
        mean_state_pred2, var_state_pred2, \
            mean_state_filt2, var_state_filt2 = jax.jit(ktv.filter)(
                mean_state_past=mean_state_past,
                var_state_past=var_state_past,
                mean_state=self.mean_state[1],
                wgt_state=self.wgt_state[0],
                var_state=self.var_state[1],
                x_meas=self.x_meas[1],
                mean_meas=self.mean_meas[1],
                wgt_meas=self.wgt_meas[1],
                var_meas=self.var_meas[1]
            )

        self.assertAlmostEqual(utils.rel_err(
            mean_state_pred1, mean_state_pred2), 0.0)
        self.assertAlmostEqual(utils.rel_err(
            var_state_pred1, var_state_pred2), 0.0)
        self.assertAlmostEqual(utils.rel_err(
            mean_state_filt1, mean_state_filt2), 0.0)
        self.assertAlmostEqual(utils.rel_err(
            var_state_filt1, var_state_filt2), 0.0)


    def test_filter2(self):
        mean_state_past, var_state_past, mean_state_pred1, var_state_pred1, \
            mean_state_filt1, var_state_filt1 = utils.test_filter2(self)
        
        mean_state_pred2, var_state_pred2, \
            mean_state_filt2, var_state_filt2 = jax.jit(ktv.filter)(
                mean_state_past=mean_state_past,
                var_state_past=var_state_past,
                mean_state=self.mean_state[2],
                wgt_state=self.wgt_state[1],
                var_state=self.var_state[2],
                x_meas=self.x_meas[2],
                mean_meas=self.mean_meas[2],
                wgt_meas=self.wgt_meas[2],
                var_meas=self.var_meas[2]
            )

        self.assertAlmostEqual(utils.rel_err(
            mean_state_pred1, mean_state_pred2), 0.0)
        self.assertAlmostEqual(utils.rel_err(
            var_state_pred1, var_state_pred2), 0.0)
        self.assertAlmostEqual(utils.rel_err(
            mean_state_filt1, mean_state_filt2), 0.0)
        self.assertAlmostEqual(utils.rel_err(
            var_state_filt1, var_state_filt2), 0.0)
        
        
    def test_smooth_mv(self):
        mean_state_next, var_state_next, mean_state_filt, var_state_filt, mean_state_pred, var_state_pred, \
            mean_state_smooth1, var_state_smooth1, _, _, _, _ = utils.test_smooth(self)
        
        mean_state_smooth2, var_state_smooth2 = jax.jit(ktv.smooth_mv)(
            mean_state_next=mean_state_next,
            var_state_next=var_state_next,
            mean_state_filt=mean_state_filt,
            var_state_filt=var_state_filt,
            mean_state_pred=mean_state_pred,
            var_state_pred=var_state_pred,
            wgt_state=self.wgt_state[0]
        )

        self.assertAlmostEqual(utils.rel_err(
            mean_state_smooth1[0], mean_state_smooth2), 0.0)
        self.assertAlmostEqual(utils.rel_err(
            var_state_smooth1[0, :, 0, :].squeeze(), var_state_smooth2), 0.0)        

        
    def test_smooth_sim(self):
        _, _, mean_state_filt, var_state_filt, mean_state_pred, var_state_pred, \
            _, _, mean_state_sim1, V, _, _ = utils.test_smooth(self)

        mean_state_sim2, var_state_sim2 = jax.jit(ktv.smooth_sim)(
            x_state_next=self.x_state_next,
            mean_state_filt=mean_state_filt,
            var_state_filt=var_state_filt,
            mean_state_pred=mean_state_pred,
            var_state_pred=var_state_pred,
            wgt_state=self.wgt_state[0]
        )
        # x_state_smooth1 = random.multivariate_normal(self.key, A.dot(self.x_state_next)+b, V)
        # x_state_smooth1 = ktv._state_sim(
        #    mean_state=A.dot(self.x_state_next)+b,
        #    var_state=V,
        #    z_state=self.z_state
        # )

        self.assertAlmostEqual(utils.rel_err(
            mean_state_sim1, mean_state_sim2), 0.0)
        self.assertAlmostEqual(utils.rel_err(V, var_state_sim2), 0.0)
        
        
    def test_smooth(self):
        mean_state_next, var_state_next, mean_state_filt, var_state_filt, mean_state_pred, var_state_pred, \
            mean_state_smooth1, var_state_smooth1, mean_state_sim1, V, _, _ = utils.test_smooth(self)

        # x_state_smooth1 = ktv._state_sim(
        #    mean_state=A.dot(self.x_state_next)+b,
        #    var_state=V,
        #    z_state=self.z_state
        # )
        # x_state_smooth1 = jax.random.multivariate_normal(self.key, A.dot(self.x_state_next)+b, V)
        mean_state_sim2, var_state_sim2, \
            mean_state_smooth2, var_state_smooth2 = jax.jit(ktv.smooth)(
                x_state_next=self.x_state_next,
                mean_state_next=mean_state_next,
                var_state_next=var_state_next,
                mean_state_filt=mean_state_filt,
                var_state_filt=var_state_filt,
                mean_state_pred=mean_state_pred,
                var_state_pred=var_state_pred,
                wgt_state=self.wgt_state[0],
            )
        self.assertAlmostEqual(utils.rel_err(
            mean_state_sim1, mean_state_sim2), 0.0)
        self.assertAlmostEqual(utils.rel_err(V, var_state_sim2), 0.0)
        self.assertAlmostEqual(utils.rel_err(
            mean_state_smooth1[0], mean_state_smooth2), 0.0)
        self.assertAlmostEqual(utils.rel_err(
            var_state_smooth1[0, :, 0, :].squeeze(), var_state_smooth2), 0.0)
        
        
    def test_smooth_cond(self):
        _, _, mean_state_filt, var_state_filt, mean_state_pred, \
            var_state_pred, _, _, _, V, A, b =  utils.test_smooth(self)
        # mean_state_sim1 = A.dot(self.x_state_next)+b
        # x_state_smooth1 = random.multivariate_normal(self.key, A.dot(self.x_state_next)+b, V)
        # x_state_smooth1 = ktv._state_sim(
        #    mean_state=A.dot(self.x_state_next)+b,
        #    var_state=V,
        #    z_state=self.z_state
        # )

        A2, b2, V2 = jax.jit(ktv.smooth_cond)(
            mean_state_filt=mean_state_filt,
            var_state_filt=var_state_filt,
            mean_state_pred=mean_state_pred,
            var_state_pred=var_state_pred,
            wgt_state=self.wgt_state[0]
        )

        self.assertAlmostEqual(utils.rel_err(A, A2), 0.0)
        self.assertAlmostEqual(utils.rel_err(b, b2), 0.0)
        self.assertAlmostEqual(utils.rel_err(V, V2), 0.0)
        


if __name__ == '__main__':
    unittest.main()



r"""
Time-varying square-root Kalman filtering and smoothing algorithms.

The Gaussian state space model underlying the algorithms is

.. math::

    x_n = c_n + Q_n x_{n-1} + R_n^{1/2} \epsilon_n

    y_n = d_n + W_n x_n + V_n^{1/2} \eta_n,

where :math:`\epsilon_n \stackrel{\text{iid}}{\sim} \operatorname{Normal}(0, I_p)` and independently :math:`\eta_n \stackrel{\text{iid}}{\sim} \operatorname{Normal}(0, I_q)`.  At each time :math:`n`, only :math:`y_n` is observed.  The suqare-root Kalman filtering and smoothing algorithms efficiently calculate quantities of the form :math:`\theta_{m|n} = (\mu_{m|n}, \Gamma_{m|n})`, where

.. math::

    \mu_{m|n} = E[x_m \mid y_{0:n}]

    \Gamma_{m|n} = \text{var}(x_m \mid y_{0:n})^{1/2},

for different combinations of :math:`m` and :math:`n`.

"""
import jax
import jax.numpy as jnp
from rodeo.utils import add_sqrt


# --- core functions -----------------------------------------------------------

def predict(mean_state_past, 
            var_state_past,
            mean_state, 
            wgt_state,
            var_state,
            *args, **kwargs):
    r"""
    Performs one prediction step of the square-root Kalman filter.

    Calculates the mean and square-root variance of :math:`p(X_n | Z_{0:n-1})` from :math:`p(X_{n-1} | Z_{0:n-1})`.

    Args:
        mean_state_past (ndarray(n_state)): State mean at time :math:`t = n-1` given observations at times :math:`t = 0, \dots, n-1`.
        var_state_past (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n-1` given observations at times :math:`t = 0, \dots, n-1`.
        mean_state (ndarray(n_state)): State equation offset at time :math:`t = n`.
        wgt_state (ndarray(n_state, n_state)): State transition matrix at time :math:`t = n`.
        var_state (ndarray(n_state, n_state)): State square-root variance matrix at time :math:`t = n`.
        args (Optional[pytree]): Additional positional arguments for Kalman functions.
        kwargs (Optional[pytree]): Additional keyword arguments for Kalman functions.

    Returns:
        (tuple):
        - **mean_state_pred** (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        - **var_state_pred** (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
    """

    mean_state_pred = wgt_state.dot(mean_state_past) + mean_state  
    var_state_pred = add_sqrt(wgt_state.dot(var_state_past), var_state)
    return mean_state_pred, var_state_pred


def update(mean_state_pred,
           var_state_pred,
           x_meas,
           mean_meas,
           wgt_meas,
           var_meas,
           *args, **kwargs):
    r"""
    Performs one update step of the square-root Kalman filter.

    Calculates the mean and square-root variance of :math:`p(X_n | Z_{0:n})` from :math:`p(X_n | Z_{0:n-1})`.

    Args:
        mean_state_pred (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        var_state_pred (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        x_meas (ndarray(n_meas)): Interrogated measure vector from :math:`x_{\text{state}}` at time :math:`t = n`.
        mean_meas (ndarray(n_meas)): Measurement equation offset at time :math:`t = n`.
        wgt_meas (ndarray(n_meas, n_state)): Measurement transition matrix at time :math:`t = n`.
        var_meas (ndarray(n_meas, n_meas)): Measurement square-root variance matrix at time :math:`t = n`.
        args (Optional[pytree]): Additional positional arguments for Kalman functions.
        kwargs (Optional[pytree]): Additional keyword arguments for Kalman functions.

    Returns:
        (tuple):
        - **mean_state_filt** (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        - **var_state_filt** (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
    """

    mean_meas_pred = wgt_meas.dot(mean_state_pred) + mean_meas
    var_meas_meas_pred = add_sqrt(wgt_meas.dot(var_state_pred),var_meas)
    # variance_state_pred = var_state_pred.dot(var_state_pred.T)
    intermediate = jax.scipy.linalg.solve_triangular(var_meas_meas_pred, wgt_meas, lower=True)
    # intermediate = intermediate.dot(variance_state_pred)
    intermediate = jnp.linalg.multi_dot([intermediate, var_state_pred, var_state_pred.T])
    var_state_temp = jax.scipy.linalg.solve_triangular(var_meas_meas_pred.T, intermediate, lower=False).T   
    mean_state_filt = mean_state_pred + \
        var_state_temp.dot(x_meas - mean_meas_pred)
    var_state_filt = add_sqrt(var_state_pred - (var_state_temp.dot(wgt_meas)).dot(var_state_pred), 
                            var_state_temp.dot(var_meas))
    
    return mean_state_filt, var_state_filt


def filter(mean_state_past,
           var_state_past,
           mean_state,
           wgt_state,
           var_state,
           x_meas,
           mean_meas,
           wgt_meas,
           var_meas,
           *args, **kwargs):
    r"""
    Performs one step of the square-root Kalman filter.

    Combines the :func:`predict` and :func:`update` steps at time :math:`t = n`.

    Args:
        mean_state_past (ndarray(n_state)): State mean at time :math:`t = n-1` given observations at times :math:`t = 0, \dots, n-1`.
        var_state_past (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n-1` given observations at times :math:`t = 0, \dots, n-1`.
        mean_state (ndarray(n_state)): State equation offset at time :math:`t = n`.
        wgt_state (ndarray(n_state, n_state)): State transition matrix at time :math:`t = n`.
        var_state (ndarray(n_state, n_state)): State square-root variance matrix at time :math:`t = n`.
        x_meas (ndarray(n_state)): Interrogated measure vector from :math:`x_{\text{state}}` at time :math:`t = n`.
        mean_meas (ndarray(n_state)): Transition offsets defining the measure prior at time :math:`t = n`.
        wgt_meas (ndarray(n_meas, n_state)): Transition matrix defining the measure prior at time :math:`t = n`.
        var_meas (ndarray(n_meas, n_meas)): Measurement square-root variance matrix at time :math:`t = n`.
        args (Optional[pytree]): Additional positional arguments for Kalman functions.
        kwargs (Optional[pytree]): Additional keyword arguments for Kalman functions.

    Returns:
        (tuple):
        - **mean_state_pred** (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        - **var_state_pred** (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        - **mean_state_filt** (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        - **var_state_filt** (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
    """

    mean_state_pred, var_state_pred = predict(
        mean_state_past=mean_state_past,
        var_state_past=var_state_past,
        mean_state=mean_state,
        wgt_state=wgt_state,
        var_state=var_state
    )
    mean_state_filt, var_state_filt = update(
        mean_state_pred=mean_state_pred,
        var_state_pred=var_state_pred,
        x_meas=x_meas,
        mean_meas=mean_meas,
        wgt_meas=wgt_meas,
        var_meas=var_meas
    )
    return mean_state_pred, var_state_pred, mean_state_filt, var_state_filt


def _smooth(var_state_filt, var_state_pred, wgt_state):
    r"""
    Common part of :func:`smooth_sim` and :func:`smooth_mv`.

    Args:
        var_state_filt (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        var_state_pred (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        wgt_state (ndarray(n_state, n_state)): State transition matrix at time :math:`t = n+1`.

    Returns:
        (ndarray(n_state, n_state)): Temporary square-root variance calculation used by :func:`smooth_sim` and :func:`smooth_mv`.
    """

    variance_state_filt = var_state_filt.dot(var_state_filt.T)
    intermediate = jax.scipy.linalg.solve_triangular(var_state_pred, wgt_state, lower=True)
    intermediate = intermediate.dot(variance_state_filt)
    var_state_temp_tilde = jax.scipy.linalg.solve_triangular(var_state_pred.T, intermediate, lower=False).T
    return var_state_temp_tilde


def smooth_mv(mean_state_next,
              var_state_next,
              mean_state_filt,
              var_state_filt,
              mean_state_pred,
              var_state_pred,
              wgt_state,
              var_state,
              *args, **kwargs):
    r"""
    Performs one step of the square-root Kalman mean/variance smoother.

    Calculates the mean and square-root variance of :math:`p(X_n | Z_{0:N})`.

    Args:
        mean_state_next (ndarray(n_state)): State mean at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, N`.
        var_state_next (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, N`.
        mean_state_filt (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        var_state_filt (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        mean_state_pred (ndarray(n_state)): State mean at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, n`.
        var_state_pred (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, n`.
        wgt_state (ndarray(n_state, n_state)): State transition matrix at time :math:`t = n+1`.
        var_state (ndarray(n_state, n_state)): State square-root variance matrix at time :math:`t = n`.
        args (Optional[pytree]): Additional positional arguments for Kalman functions.
        kwargs (Optional[pytree]): Additional keyword arguments for Kalman functions.

    Returns:
        (tuple):
        - **mean_state_smooth** (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
        - **var_state_smooth** (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
    """

    var_state_temp_tilde = _smooth(
        var_state_filt, var_state_pred, wgt_state
    )
    mean_state_smooth = mean_state_filt + \
        var_state_temp_tilde.dot(mean_state_next - mean_state_pred)
    I = jnp.eye(var_state_temp_tilde.shape[0])    
    J = I - jnp.matmul(var_state_temp_tilde, wgt_state)
    var_state_smooth = add_sqrt(jnp.matmul(var_state_temp_tilde, jnp.hstack((var_state_next, var_state))),
                                jnp.matmul(J,var_state_filt))
    return mean_state_smooth, var_state_smooth


def smooth_sim(x_state_next,
               mean_state_filt,
               var_state_filt,
               mean_state_pred,
               var_state_pred,
               wgt_state,
               var_state,
               *args, **kwargs):
    r"""
    Performs one step of the square-root Kalman sampling smoother.

    Calculates the mean and square-root variance of :math:`p(X_n | X_{n+1}, Z_{0:N})`.

    Args:
        x_state_next (ndarray(n_state)): Simulated state at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, N`.
        mean_state_filt (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        var_state_filt (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        mean_state_pred (ndarray(n_state)): State mean at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, n`.
        var_state_pred (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, n`.
        wgt_state (ndarray(n_state, n_state)): State transition matrix at time :math:`t = n+1`.
        var_state (ndarray(n_state, n_state)): State square-root variance matrix at time :math:`t = n`.
        args (Optional[pytree]): Additional positional arguments for Kalman functions.
        kwargs (Optional[pytree]): Additional keyword arguments for Kalman functions.

    Returns:
        (tuple):
        - **mean_state_sim** (ndarray(n_state)): Mean for the sample solution at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
        - **var_state_sim** (ndarray(n_state)): Square-root variance for the sample solution at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
    """

    var_state_temp_tilde = _smooth(
        var_state_filt, var_state_pred, wgt_state
    )
    mean_state_sim = mean_state_filt + \
        var_state_temp_tilde.dot(x_state_next - mean_state_pred)
    I = jnp.eye(var_state_temp_tilde.shape[0])    
    J = I - jnp.matmul(var_state_temp_tilde, wgt_state)
    var_state_sim = add_sqrt(jnp.matmul(var_state_temp_tilde, var_state),
                             jnp.matmul(J,var_state_filt))
    return mean_state_sim, var_state_sim


def smooth(x_state_next,
           mean_state_next,
           var_state_next,
           mean_state_filt,
           var_state_filt,
           mean_state_pred,
           var_state_pred,
           wgt_state,
           var_state,
           *args, **kwargs):
    r"""
    Performs one step of both square-root Kalman mean/variance and sampling smoothers.

    Args:
        x_state_next (ndarray(n_state)): Simulated state at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, N`.
        mean_state_next (ndarray(n_state)): State mean at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, N`.
        var_state_next (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, N`.
        mean_state_filt (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        var_state_filt (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        mean_state_pred (ndarray(n_state)): State mean at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, n`.
        var_state_pred (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, n`.
        wgt_state (ndarray(n_state, n_state)): State transition matrix at time :math:`t = n+1`.
        var_state (ndarray(n_state, n_state)): State square-root variance matrix at time :math:`t = n`.
        args (Optional[pytree]): Additional positional arguments for Kalman functions.
        kwargs (Optional[pytree]): Additional keyword arguments for Kalman functions.

    Returns:
        (tuple):
        - **mean_state_sim** (ndarray(n_state)): Mean for the sample solution at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
        - **var_state_sim** (ndarray(n_state)): Square-root variance for the sample solution at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
        - **mean_state_smooth** (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
        - **var_state_smooth** (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
    """

    var_state_temp_tilde = _smooth(
        var_state_filt, var_state_pred, wgt_state
    )
    mean_state_temp = jnp.concatenate([x_state_next[None],
                                       mean_state_next[None]])
    mean_state_temp = mean_state_filt + \
        var_state_temp_tilde.dot((mean_state_temp - mean_state_pred).T).T
    mean_state_sim = mean_state_temp[0]
    mean_state_smooth = mean_state_temp[1]
    I = jnp.eye(var_state_temp_tilde.shape[0])    
    J = I - jnp.matmul(var_state_temp_tilde, wgt_state)
    var_state_sim = add_sqrt(jnp.matmul(var_state_temp_tilde, var_state),
                             jnp.matmul(J,var_state_filt))
    var_state_smooth = add_sqrt(jnp.matmul(var_state_temp_tilde, 
                                           jnp.hstack((var_state_next, var_state))),
                                jnp.matmul(J,var_state_filt))
    return mean_state_sim, var_state_sim, mean_state_smooth, var_state_smooth


def forecast(mean_state_pred,
             var_state_pred,
             mean_meas,
             wgt_meas,
             var_meas,
             *args, **kwargs):
    r"""
    Forecasts the mean and variance of the measurement at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.

    Args:
        mean_state_pred (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        var_state_pred (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        mean_meas (ndarray(n_meas)): Measurement equation offset at time :math:`t = n`.
        wgt_meas (ndarray(n_meas, n_state)): Measurement transition matrix at time :math:`t = n`.
        var_meas (ndarray(n_meas, n_meas)): State square-root variance matrix at time :math:`t = n`.
        args (Optional[pytree]): Additional positional arguments for Kalman functions.
        kwargs (Optional[pytree]): Additional keyword arguments for Kalman functions.

    Returns:
        (tuple):
        - **mean_fore** (ndarray(n_meas)): Forecast mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
        - **var_fore** (ndarray(n_meas, n_meas)): Forecast variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n-1`.
    """

    # wgt_meas = W + wgt_meas
    mean_fore = wgt_meas.dot(mean_state_pred) + mean_meas
    var_fore = add_sqrt(wgt_meas.dot(var_state_pred), var_meas)   
    var_fore = var_fore.dot(var_fore.T)
    return mean_fore, var_fore


def smooth_cond(mean_state_filt,
                var_state_filt,
                mean_state_pred,
                var_state_pred,
                wgt_state,
                var_state,
                *args, **kwargs):
    r"""
    Performs one step of the square-root Kalman sampling smoother conditional.

    Finds :math:`A_n`, :math:`b_n`, and :math:`C_n = \text{cholesky}(V_n)` such that :math:`X_n \sim \mathcal{N}(A_n X_{n+1} + b_n, V_n)`. This is similar to the Kalman sampling algorithm but without the explicit sample.

    Args:
        mean_state_filt (ndarray(n_state)): State mean at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        var_state_filt (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n` given observations at times :math:`t = 0, \dots, n`.
        mean_state_pred (ndarray(n_state)): State mean at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, n`.
        var_state_pred (ndarray(n_state, n_state)): State square-root variance at time :math:`t = n+1` given observations at times :math:`t = 0, \dots, n`.
        wgt_state (ndarray(n_state, n_state)): State transition matrix at time :math:`t = n+1`.
        var_state (ndarray(n_state, n_state)): State square-root variance matrix at time :math:`t = n`.
        args (Optional[pytree]): Additional positional arguments for Kalman functions.
        kwargs (Optional[pytree]): Additional keyword arguments for Kalman functions.

    Returns:
        (tuple):
        - **wgt_state_cond** (ndarray(n_state, n_state)): Transition of smooth conditional at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
        - **mean_state_cond** (ndarray(n_state)): Offset of smooth conditional at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
        - **var_state_cond** (ndarray(n_state, n_state)): Square-root variance of smooth conditional at time :math:`t = n` given observations at times :math:`t = 0, \dots, N`.
    """

    wgt_state_cond = _smooth(
        var_state_filt, var_state_pred, wgt_state
    )
    mean_state_cond = mean_state_filt - wgt_state_cond.dot(mean_state_pred)
    I = jnp.eye(wgt_state_cond.shape[0])    
    J = I - jnp.matmul(wgt_state_cond, wgt_state)
    var_state_cond = add_sqrt(jnp.matmul(wgt_state_cond, var_state),
                             jnp.matmul(J,var_state_filt))
    return wgt_state_cond, mean_state_cond, var_state_cond





"""
Implements the user interfaces for Random Walk Rosenbluth-Metropolis-Hastings kernels with auxiliary variables.
That is, the `logdensity_fn` returns a tuple of which the first element is the log-density, and the second
is a PyTree containing the auxiliary variables.
The remainder of this file is nearly identical to `blackjax.mcmc.random_walk.py`. 

Some interfaces are exposed here for convenience and for entry level users, who might be familiar
with simpler versions of the algorithms, but in all cases they are particular instantiations
of the Random Walk Rosenbluth-Metropolis-Hastings.

Let's note :math:`x_{t-1}` to the previous position and :math:`x_t` to the newly sampled one.

The variants offered are:

1. Proposal distribution as addition of random noice from previous position. This means
   `x_t = x_{t-1} + step`.

    Function: :func:`additive_step`

2. Independent proposal distribution: :math:`P(x_t)` doesn't depend on :math:`x_{t_1}`.

    Function: :func:`irmh`

3. Proposal distribution using a symmetric function. That means :math:`P(x_t|x_{t-1}) = P(x_{t-1}|x_t)`.
   See 'Metropolis Algorithm' in [1].

    Function: :func:`rmh` without proposal_logdensity_fn.

4. Asymmetric proposal distribution. See 'Metropolis-Hastings' Algorithm in [1].

    Function: :func:`rmh` with proposal_logdensity_fn.

    
Reference: Andrew Gelman, John B Carlin, Hal S Stern, and Donald B Rubin. Bayesian data analysis. Chapman and Hall/CRC, 2014. Section 11.2

Example:    
    The simplest case is:

    .. code::

        random_walk = blackjax.additive_step_random_walk(logdensity_fn, blackjax.mcmc.random_walk.normal(sigma))
        state = random_walk.init(position)
        new_state, info = random_walk.step(rng_key, state)

    In all cases we can JIT-compile the step function for better performance

    .. code::

        step = jax.jit(random_walk.step)
        new_state, info = step(rng_key, state)

"""
from typing import Callable, NamedTuple, Optional

import jax
from jax import numpy as jnp

from blackjax.base import SamplingAlgorithm
from blackjax.mcmc import proposal
from blackjax.types import Array, ArrayLikeTree, ArrayTree, PRNGKey
from blackjax.mcmc.random_walk import normal as normal

# ignore this for now
# __all__ = [
#     "build_additive_step",
#     "normal",
#     "build_irmh",
#     "build_rmh",
#     "RWInfo",
#     "RWState",
#     "rmh_proposal",
#     "build_rmh_transition_energy",
#     "additive_step_random_walk",
#     "irmh_as_top_level_api",
#     "rmh_as_top_level_api",
#     "normal_random_walk",
# ]


# imported from blackjax.mcmc.random_walk
# def normal(sigma: Array) -> Callable:
#     """Normal Random Walk proposal.

#     Propose a new position such that its distance to the current position is
#     normally distributed. Suitable for continuous variables.

#     Parameter
#     ---------
#     sigma:
#         vector or matrix that contains the standard deviation of the centered
#         normal distribution from which we draw the move proposals.

#     """
#     if jnp.ndim(sigma) > 2:
#         raise ValueError("sigma must be a vector or a matrix.")

#     def propose(rng_key: PRNGKey, position: ArrayLikeTree) -> ArrayTree:
#         return generate_gaussian_noise(rng_key, position, sigma=sigma)

#     return propose


class RWAState(NamedTuple):
    """
    State of the Random Walk Auxiliary (RWA) chain.

    """

    position: ArrayTree
    """Current position of the chain."""

    logdensity: float
    """Current value of the log-density."""

    auxdata: ArrayTree = None
    """Current value of the auxiliary data."""


class RWAInfo(NamedTuple):
    """
    Additional information about the RWA chain step.

    """

    acceptance_rate: float
    """acceptance_rate (float): The acceptance probability of the proposed transition."""
    
    is_accepted: bool
    """Indicates whether the proposed state was accepted."""

    proposal: RWAState
    """The proposed state of the chain."""


def init(
    position: ArrayLikeTree, logdensity_fn: Callable, rng_key: PRNGKey
) -> RWAState:
    """
    Create an initial chain state from a given position.

    Args:
        position (ArrayLikeTree): The initial position of the chain.
        logdensity_fn (Callable): Function to compute the log-probability density of the distribution.

    Returns:
        (RWAState): The initialized state of the chain.
    """
    logdensity, auxdata = logdensity_fn(position, rng_key)
    return RWAState(position, logdensity, auxdata)


def build_additive_step():
    """
    Build a Random Walk Rosenbluth-Metropolis-Hastings (RMH) kernel using an additive step proposal.

    Returns:
        (Callable): A function that takes a random key and chain state, performs an RMH step, and returns the new state and transition info.
    """

    def kernel(
        rng_key: PRNGKey, state: RWAState, logdensity_fn: Callable, random_step: Callable
    ) -> tuple[RWAState, RWAInfo]:
        def proposal_generator(key_proposal, position):
            move_proposal = random_step(key_proposal, position)
            new_position = jax.tree_util.tree_map(
                jnp.add, position, move_proposal)
            return new_position

        inner_kernel = build_rmh()
        return inner_kernel(rng_key, state, logdensity_fn, proposal_generator)

    return kernel


def normal_random_walk(logdensity_fn: Callable, sigma):
    """
    Create a Gaussian additive step random walk Metropolis-Hastings sampler.

    This method initializes a random walk sampler with Gaussian-distributed steps.

    Args:
        logdensity_fn (Callable): Function to compute the log-probability density of the distribution.
        sigma (ArrayLikeTree): Standard deviation of the Gaussian distribution used for the proposal steps.

    Returns:
        (SamplingAlgorithm): An object with `init` and `step` methods to run the Gaussian RMH sampler.

    """
    return additive_step_random_walk(logdensity_fn, normal(sigma))


def additive_step_random_walk(
    logdensity_fn: Callable, random_step: Callable
) -> SamplingAlgorithm:
    """Implements the user interface for the Additive Step RMH

    Example:
        A new kernel can be initialized and used with the following code:

        .. code::

            rw = blackjax.additive_step_random_walk(logdensity_fn, random_step)
            state = rw.init(position)
            new_state, info = rw.step(rng_key, state)

        The specific case of a Gaussian `random_step` is already implemented, either with independent components
        when `covariance_matrix` is a one dimensional array or with dependent components if a two dimensional array:

        .. code::

            rw_gaussian = blackjax.additive_step_random_walk.normal_random_walk(logdensity_fn, covariance_matrix)
            state = rw_gaussian.init(position)
            new_state, info = rw_gaussian.step(rng_key, state)

    Args:
        logdensity_fn (Callable): Function to compute the log-probability density of the distribution.
        random_step (Callable): A function that generates a step to be added to the current state. This
            function takes a PRNG key and the current position as input and returns a new proposal step.

    Returns:
        (SamplingAlgorithm): A sampling algorithm with `init` and `step` methods to perform RMH sampling.
    """
    kernel = build_additive_step()

    def init_fn(position: ArrayLikeTree, rng_key=None):
        # del rng_key
        return init(position, logdensity_fn, rng_key)

    def step_fn(rng_key: PRNGKey, state):
        return kernel(rng_key, state, logdensity_fn, random_step)

    return SamplingAlgorithm(init_fn, step_fn)


def build_irmh() -> Callable:
    """
    Build an Independent Random Walk Rosenbluth-Metropolis-Hastings (RMH) kernel.

    This kernel uses a proposal distribution that is independent of the current state, i.e., 
    the new proposed state is sampled independently of the particle's current position.

    Returns:
        (Callable): A function (kernel) that takes a PRNG key and a PyTree containing the 
        current state of the chain and that returns a new state of the chain along with
        information about the transition.
    """

    def kernel(
        rng_key: PRNGKey,
        state: RWAState,
        logdensity_fn: Callable,
        proposal_distribution: Callable,
        proposal_logdensity_fn: Optional[Callable] = None,
    ) -> tuple[RWAState, RWAInfo]:
        """
        Args:
            proposal_distribution (Callable): A function that takes a PRNG key and returns a 
                sample in the same domain as the target distribution.
            proposal_logdensity_fn (Optional[Callable]): A function that returns the log-density
                of obtaining a given proposal, given the current state. This is required
                for non-symmetric proposals where P(x_t | x_{t-1}) ≠ P(x_{t-1} | x_t). If not
                provided, the proposal is assumed to be symmetric.
        """

        def proposal_generator(rng_key: PRNGKey, position: ArrayTree):
            del position
            return proposal_distribution(rng_key)

        inner_kernel = build_rmh()
        return inner_kernel(
            rng_key, state, logdensity_fn, proposal_generator, proposal_logdensity_fn
        )

    return kernel


def irmh_as_top_level_api(
    logdensity_fn: Callable,
    proposal_distribution: Callable,
    proposal_logdensity_fn: Optional[Callable] = None,
) -> SamplingAlgorithm:
    """
    Implements the (basic) user interface for the independent RMH.

    Example:
        A new kernel can be initialized and used with the following code:

        .. code::

            rmh = blackjax.irmh(logdensity_fn, proposal_distribution)
            state = rmh.init(position)
            new_state, info = rmh.step(rng_key, state)

        We can JIT-compile the step function for better performance

        .. code::

            step = jax.jit(rmh.step)
            new_state, info = step(rng_key, state)

    Args:
        logdensity_fn (Callable): The log-probability density function of the distribution to sample from.
        proposal_distribution (Callable): A function that takes a PRNG key and produces a new proposal.
            The proposal is independent of the current state of the sampler.
        proposal_logdensity_fn (Optional[Callable]): A function that returns the log-density of obtaining
            a given proposal, given the current state. This is required for non-symmetric proposals.
            If not provided, the proposal is assumed to be symmetric.

    Returns:
        (SamplingAlgorithm): An object containing `init` and `step` methods for performing
        Independent Random Walk Metropolis-Hastings sampling.

    """
    kernel = build_irmh()

    def init_fn(position: ArrayLikeTree, rng_key=None):
        # del rng_key
        return init(position, logdensity_fn, rng_key)

    def step_fn(rng_key: PRNGKey, state):
        return kernel(
            rng_key,
            state,
            logdensity_fn,
            proposal_distribution,
            proposal_logdensity_fn,
        )

    return SamplingAlgorithm(init_fn, step_fn)


def build_rmh():
    """Build a Rosenbluth-Metropolis-Hastings kernel.

    Returns:
        (Callable): A function (kernel) that takes a PRNG key and a PyTree containing the 
        current state of the chain and that returns a new state of the chain along with
        information about the transition.

    """

    def kernel(
        rng_key: PRNGKey,
        state: RWAState,
        logdensity_fn: Callable,
        transition_generator: Callable,
        proposal_logdensity_fn: Optional[Callable] = None,
    ) -> tuple[RWAState, RWAInfo]:
        """
        Perform one step of the Rosenbluth-Metropolis-Hastings (RMH) algorithm.

        This function moves the Markov chain by one step using the RMH algorithm by generating a candidate state and accepting or rejecting it based on the Metropolis-Hastings acceptance criterion.

        Args:
            rng_key (PRNGKey): The pseudo-random number generator key used to generate random numbers.
            state (RWAState): The current state of the Markov chain.
            logdensity_fn (Callable): A function that computes the log-probability at a given position.
            transition_generator (Callable): A function that generates a candidate transition for the Markov chain.
            proposal_logdensity_fn (Optional[Callable]): A function that returns the log-density of obtaining a given proposal from the current state. Required for non-symmetric proposals. If not provided, the proposal is assumed to be symmetric.

        Returns:
            (tuple):
                - **RWAState**: The new state of the Markov chain after the RMH step.
                - **RWAInfo**: Additional information about the step, such as the acceptance probability and whether the proposal was accepted.

        """
        transition_energy = build_rmh_transition_energy(proposal_logdensity_fn)

        compute_acceptance_ratio = proposal.compute_asymmetric_acceptance_ratio(
            transition_energy
        )

        proposal_generator = rmh_proposal(
            logdensity_fn, transition_generator, compute_acceptance_ratio
        )
        new_state, do_accept, p_accept = proposal_generator(rng_key, state)
        return new_state, RWAInfo(p_accept, do_accept, new_state)

    return kernel


def rmh_as_top_level_api(
    logdensity_fn: Callable,
    proposal_generator: Callable[[PRNGKey, ArrayLikeTree], ArrayTree],
    proposal_logdensity_fn: Optional[Callable[[
        ArrayLikeTree], ArrayTree]] = None,
) -> SamplingAlgorithm:
    """
    Implements the user interface for the RMH.

    Example:
        A new kernel can be initialized and used with the following code:

        .. code::

            rmh = blackjax.rmh(logdensity_fn, proposal_generator)
            state = rmh.init(position)
            new_state, info = rmh.step(rng_key, state)

        We can JIT-compile the step function for better performance

        .. code::

            step = jax.jit(rmh.step)
            new_state, info = step(rng_key, state)

        Create a user interface for the Rosenbluth-Metropolis-Hastings (RMH) sampler.

        This function returns a `SamplingAlgorithm` object that provides `init` and `step` methods for performing RMH sampling. The user can specify a custom proposal generator and an optional log-density function for non-symmetric proposals.

    Args:
        logdensity_fn (Callable): The log-probability density function of the distribution to sample from.
        proposal_generator (Callable): A function that takes a random number generator key and the current state, then generates a new proposal.
        proposal_logdensity_fn (Optional[Callable]): The log-density function associated with the proposal generator. If the proposal distribution is non-symmetric (i.e., P(x_t | x_{t-1}) ≠ P(x_{t-1} | x_t)), this must be provided to apply the Metropolis-Hastings correction for detailed balance.

    Returns:
        (SamplingAlgorithm): An object containing `init` and `step` methods for running the RMH sampler.

    """
    kernel = build_rmh()

    def init_fn(position: ArrayLikeTree, rng_key=None):
        # del rng_key
        return init(position, logdensity_fn, rng_key)

    def step_fn(rng_key: PRNGKey, state):
        return kernel(
            rng_key,
            state,
            logdensity_fn,
            proposal_generator,
            proposal_logdensity_fn,
        )

    return SamplingAlgorithm(init_fn, step_fn)


def build_rmh_transition_energy(proposal_logdensity_fn: Optional[Callable]) -> Callable:
    if proposal_logdensity_fn is None:

        def transition_energy(prev_state, new_state):
            return -new_state.logdensity

    else:

        def transition_energy(prev_state, new_state):
            return -new_state.logdensity - proposal_logdensity_fn(new_state, prev_state)

    return transition_energy


def rmh_proposal(
    logdensity_fn: Callable,
    transition_distribution: Callable,
    compute_acceptance_ratio: Callable,
    sample_proposal: Callable = proposal.static_binomial_sampling,
) -> Callable:
    """
    Args:
        logdensity_fn (Callable): The log-probability density function of the distribution to sample from.
        transition_distribution (Callable): A function that takes a random number generator key and the current state, then generates a new proposal.
        compute_acceptance_ratio (Callable): A function to compute the acceptance ratio.
        sample_proposal (Callable): A function to generate the next sample given proposal and previous state.
    
    Returns:
        (Callable): Generator for sample proposals.
    """

    def generate(rng_key, previous_state: RWAState) -> tuple[RWAState, bool, float]:
        key_proposal, key_accept, key_logdensity = jax.random.split(rng_key, 3)
        position, _, _ = previous_state
        new_position = transition_distribution(key_proposal, position)
        new_logdensity, new_auxdata = logdensity_fn(new_position, key_logdensity)
        proposed_state = RWAState(new_position, new_logdensity, new_auxdata)
        log_p_accept = compute_acceptance_ratio(previous_state, proposed_state)
        accepted_state, info = sample_proposal(
            key_accept, log_p_accept, previous_state, proposed_state
        )
        do_accept, p_accept, _ = info
        return accepted_state, do_accept, p_accept


    return generate



import unittest
import jax
from rodeo.solve import *
from rodeo.interrogate import *
import ode_block_solve_for as bfor
import utils
# from jax import config
# config.update("jax_enable_x64", True)

class TestrodeoFor(unittest.TestCase):
    """
    Test if lax scan version of rodeo gives the same results as for-loop version.

    """
    setUp = utils.fitz_setup 

    def test_interrogate_rodeo(self):
        wgt_meas1, mean_meas1, var_meas1 = interrogate_rodeo(
            key=self.key,
            ode_fun=self.fitz_jax,
            ode_weight=self.W_block,
            t=self.t,
            theta=self.theta,
            mean_state_pred=self.x0_block,
            var_state_pred=self.prior_R
        )
        # for
        wgt_meas2, mean_meas2, var_meas2 = bfor.interrogate_rodeo(
            key=self.key,
            ode_fun=self.fitz_jax,
            ode_weight=self.W_block,
            t=self.t,
            theta=self.theta,
            mean_state_pred=self.x0_block,
            var_state_pred=self.prior_R
        )
        
        self.assertAlmostEqual(utils.rel_err(wgt_meas1, wgt_meas2), 0.0)
        self.assertAlmostEqual(utils.rel_err(mean_meas1, mean_meas2), 0.0)
        self.assertAlmostEqual(utils.rel_err(var_meas1, var_meas2), 0.0)
    
    def test_interrogate_chkrebtii(self):
        wgt_meas1, mean_meas1, var_meas1 = interrogate_chkrebtii(
            key=self.key,
            ode_fun=self.fitz_jax,
            ode_weight=self.W_block,
            t=self.t,
            theta=self.theta,
            mean_state_pred=self.x0_block,
            var_state_pred=self.prior_R,
            kalman_type="standard"
        )
        # for
        wgt_meas2, mean_meas2, var_meas2 = bfor.interrogate_chkrebtii(
            key=self.key,
            ode_fun=self.fitz_jax,
            ode_weight=self.W_block,
            t=self.t,
            theta=self.theta,
            mean_state_pred=self.x0_block,
            var_state_pred=self.prior_R
        )
        
        self.assertAlmostEqual(utils.rel_err(wgt_meas1, wgt_meas2), 0.0)
        self.assertAlmostEqual(utils.rel_err(mean_meas1, mean_meas2), 0.0)
        self.assertAlmostEqual(utils.rel_err(var_meas1, var_meas2), 0.0)

    def test_interrogate_kramer(self):
        wgt_meas1, mean_meas1, var_meas1 = interrogate_kramer(
            key=self.key,
            ode_fun=self.fitz_jax,
            ode_weight=self.W_block,
            t=self.t,
            theta=self.theta,
            mean_state_pred=self.x0_block,
            var_state_pred=self.prior_R
        )
        # for
        wgt_meas2, mean_meas2, var_meas2 = bfor.interrogate_kramer(
            key=self.key,
            ode_fun=self.fitz_jax,
            ode_weight=self.W_block,
            t=self.t,
            theta=self.theta,
            mean_state_pred=self.x0_block,
            var_state_pred=self.prior_R
        )
        
        self.assertAlmostEqual(utils.rel_err(wgt_meas1, wgt_meas2), 0.0)
        self.assertAlmostEqual(utils.rel_err(mean_meas1, mean_meas2), 0.0)
        self.assertAlmostEqual(utils.rel_err(var_meas1, var_meas2), 0.0)

    def test_solve_sim(self):
        sim1 = solve_sim(key=self.key, ode_fun=self.fitz_jax, ode_weight=self.W_block,
                         ode_init=self.x0_block, theta=self.theta,
                         t_min=self.t_min, t_max=self.t_max, n_steps=self.n_steps, 
                         prior_pars=self.prior_pars,
                         interrogate=interrogate_rodeo)
        # for
        sim2 = bfor.solve_sim(key=self.key, ode_fun=self.fitz_jax, ode_weight=self.W_block,
                              ode_init=self.x0_block, theta=self.theta,
                              t_min=self.t_min, t_max=self.t_max, n_steps=self.n_steps, 
                              prior_pars=self.prior_pars,
                              interrogate=interrogate_rodeo)
        
        self.assertAlmostEqual(utils.rel_err(sim1, sim2), 0.0)
    
    def test_solve_mv(self):
        mu1, var1 = solve_mv(key=self.key, ode_fun=self.fitz_jax, ode_weight=self.W_block,
                             ode_init=self.x0_block, theta=self.theta,
                             t_min=self.t_min, t_max=self.t_max, n_steps=self.n_steps, 
                             prior_pars=self.prior_pars,
                             interrogate=interrogate_rodeo)
        # for
        mu2, var2 = bfor.solve_mv(key=self.key, ode_fun=self.fitz_jax, ode_weight=self.W_block,
                                  ode_init=self.x0_block, theta=self.theta,
                                  t_min=self.t_min, t_max=self.t_max, n_steps=self.n_steps, 
                                  prior_pars=self.prior_pars,
                                  interrogate=interrogate_rodeo)
        self.assertAlmostEqual(utils.rel_err(mu1, mu2), 0.0)
        self.assertAlmostEqual(utils.rel_err(var1, var2), 0.0)
    
    # def test_solve(self):
    #     sim1, mu1, var1 = solve(key=self.key, ode_fun=self.fitz_jax, ode_weight=self.W_block,
    #                             ode_init=self.x0_block, theta=self.theta,
    #                             t_min=self.t_min, t_max=self.t_max, n_steps=self.n_steps, 
    #                             prior_weight=self.prior_Q, prior_var=self.prior_R,
    #                             interrogate=interrogate_kramer)
    #     # for
    #     sim2, mu2, var2 = bfor.solve(key=self.key, ode_fun=self.fitz_jax, ode_weight=self.W_block,
    #                                  ode_init=self.x0_block, theta=self.theta,
    #                                  t_min=self.t_min, t_max=self.t_max, n_steps=self.n_steps, 
    #                                  prior_weight=self.prior_Q, prior_var=self.prior_R,
    #                                  interrogate=interrogate_kramer)
    #     self.assertAlmostEqual(utils.rel_err(mu1, mu2), 0.0)
    #     self.assertAlmostEqual(utils.rel_err(var1, var2), 0.0)
    #     self.assertAlmostEqual(utils.rel_err(sim1, sim2), 0.0, places=7)

if __name__ == '__main__':
    unittest.main()
